@@ -1,3 +1,5 @@
+//! Types for attaching to processes, managing tracees, and interpreting tracing events.
+
 use std::collections::BTreeMap;
 use std::fs;
 use std::io;
@@ -10,45 +12,63 @@ use nix::sys::{
     wait::{self, WaitPidFlag, WaitStatus},
 };
 
-use crate::error::{Error, Result};
+use crate::error::{Error, Result, ResultExt};
 
+#[cfg(target_arch = "x86_64")]
+use crate::x86::DebugRegister;
 
 pub use nix::unistd::Pid;
 pub use nix::sys::ptrace::Options;
+
+/// POSIX signal.
 pub use nix::sys::signal::Signal;
 
+/// Register state of a tracee.
+#[cfg(target_arch = "x86_64")]
 pub type Registers = libc::user_regs_struct;
+
+/// Extra signal info, such as its cause.
 pub type Siginfo = libc::siginfo_t;
 
 const WALL: Option<WaitPidFlag> = Some(WaitPidFlag::__WALL);
 
-/// Various ptrace-stops.
+/// A _ptrace-stop_, a tracee state in which it is stopped and ready to accept ptrace
+/// commands.
+///
+/// These ptrace-stops may carry data obtained via additional (internal) ptrace requests
+/// to `PTRACE_GETEVENTMSG`. Requests to `PTRACE_GETSIGINFO` may be made to disambiguate
+/// stops.
 #[derive(Clone, Copy, Debug, Eq, PartialEq)]
 pub enum Stop {
-    AttachStop(Pid),
+    Attach,
 
     // signal-delivery-stop
-    SignalDeliveryStop(Pid, Signal),
+    SignalDelivery { signal: Signal },
 
     // group-stop
-    GroupStop(Pid, Signal),
+    Group { signal: Signal },
 
     // syscall-stops
-    SyscallEnterStop(Pid),
-    SyscallExitStop(Pid),
+    SyscallEnter,
+    SyscallExit,
 
     // ptrace-event-stops
-    Clone(Pid, Pid),
-    Fork(Pid, Pid),
-    Exec(Pid, Pid),
-    Exiting(Pid, i32),
-    Signaling(Pid, Signal, bool),
-    Vfork(Pid, Pid),
-    VforkDone(Pid, Pid),
-    Seccomp(u16),
+    Clone { new: Pid },
+    Fork { new: Pid },
+    Exec { old: Pid },
+    Exiting { exit_code: i32 },
+    Signaling {
+        signal: Signal,
+        core_dumped: bool,
+    },
+    Vfork { new: Pid },
+    VforkDone { new: Pid },
+    Seccomp { data: u16 },
 }
 
-/// Ptrace restart requests.
+/// Restart requests, which resume stopped tracees.
+///
+/// The restart mode determines the possible subsequent stops of the restarted tracee.
 #[derive(Clone, Copy, Debug, Eq, PartialEq)]
 pub enum Restart {
     Step,
@@ -56,19 +76,18 @@ pub enum Restart {
     Syscall,
 }
 
-/// Tracee in ptrace-stop, with an optional pending signal.
+/// Tracee task in ptrace-stop, with an optional pending signal.
 ///
-/// Describes how the stopped tracee would continue if it weren't traced, and thus how to
-/// restart it to resume normal execution.
-///
-/// The underlying tracee is not guaranteed to exist.
+/// **Warning:** the underlying tracee is not guaranteed to exist, and
+/// operations on it may fail between calls to [`Ptracer::wait()`].
 #[derive(Clone, Copy, Debug, Eq, PartialEq)]
 pub struct Tracee {
     pub pid: Pid,
     pub pending: Option<Signal>,
     pub stop: Stop,
 
-    pub _not_send: PhantomData<*const ()>,
+    #[doc(hidden)]
+    _not_send: PhantomData<*const ()>,
 }
 
 impl Tracee {
@@ -91,15 +110,17 @@ impl Tracee {
 
     /// Set custom tracing options on the tracee.
     pub fn set_options(&mut self, options: Options) -> Result<()> {
-        Ok(ptrace::setoptions(self.pid, options)?)
+        Ok(ptrace::setoptions(self.pid, options).died_if_esrch(self.pid)?)
     }
 
+    #[cfg(target_arch = "x86_64")]
     pub fn registers(&self) -> Result<Registers> {
-        Ok(ptrace::getregs(self.pid)?)
+        Ok(ptrace::getregs(self.pid).died_if_esrch(self.pid)?)
     }
 
+    #[cfg(target_arch = "x86_64")]
     pub fn set_registers(&mut self, regs: Registers) -> Result<()> {
-        Ok(ptrace::setregs(self.pid, regs)?)
+        Ok(ptrace::setregs(self.pid, regs).died_if_esrch(self.pid)?)
     }
 
     pub fn read_memory(&mut self, addr: u64, len: usize) -> Result<Vec<u8>> {
@@ -137,14 +158,74 @@ impl Tracee {
     }
 
     pub fn siginfo(&self) -> Result<Option<Siginfo>> {
-        let info = if let Stop::SignalDeliveryStop(..) = self.stop {
-            Some(ptrace::getsiginfo(self.pid)?)
+        let info = if let Stop::SignalDelivery { .. } = self.stop {
+            Some(ptrace::getsiginfo(self.pid).died_if_esrch(self.pid)?)
         } else {
             None
         };
 
         Ok(info)
     }
+
+    #[cfg(target_arch = "x86_64")]
+    pub fn debug_register(&self, dr: DebugRegister) -> Result<u64> {
+        let index = 8 * u64::from(dr);
+
+        if let Some(off) = DebugRegister::user_offset().checked_add(index) {
+            self.peek_user(off)
+        } else {
+            internal_error!("unreachable overflow")
+        }
+    }
+
+    #[cfg(target_arch = "x86_64")]
+    pub fn set_debug_register(&self, dr: DebugRegister, data: u64) -> Result<()> {
+        let index = 8 * u64::from(dr);
+
+        if let Some(off) = DebugRegister::user_offset().checked_add(index) {
+            self.poke_user(off, data)
+        } else {
+            internal_error!("unreachable overflow")
+        }
+    }
+
+    #[cfg(target_arch = "x86_64")]
+    fn peek_user(&self, off: u64) -> Result<u64> {
+        // SAFETY: `off` does not require validation, because it is not actually used as a
+        // pointer offset by the kernel.
+        //
+        // See: https://github.com/torvalds/linux/blob/v4.9/arch/x86/kernel/ptrace.c#L774-L791
+
+        let data = unsafe {
+            libc::ptrace(
+                libc::PTRACE_PEEKUSER,
+                self.pid,
+                off,
+                0,
+            )
+        };
+
+        Ok(data as u64)
+    }
+
+    #[cfg(target_arch = "x86_64")]
+    fn poke_user(&self, off: u64, data: u64) -> Result<()> {
+        // SAFETY: `off` does not require validation, because it is not actually used as a
+        // pointer offset by the kernel.
+        //
+        // See: https://github.com/torvalds/linux/blob/v4.9/arch/x86/kernel/ptrace.c#L774-L791
+
+        unsafe {
+            libc::ptrace(
+                libc::PTRACE_POKEUSER,
+                self.pid,
+                off,
+                data,
+            )
+        };
+
+        Ok(())
+    }
 }
 
 #[derive(Clone, Copy, Debug, Eq, PartialEq)]
@@ -162,7 +243,12 @@ enum State {
     Syscalling,
 }
 
-/// Tracer for a (possibly multi-threaded) Linux process.
+/// Tracer for a Linux process.
+///
+/// By default, [spawning](Ptracer::spawn()) a child tracee will follow calls to `fork()`,
+/// `clone()`, and `exec()`, tracing any child tasks (both threads and processes).
+///
+/// This can be configured for any stopped tracee via [`Tracee::set_options()`].
 #[derive(Clone, Debug, Eq, PartialEq)]
 pub struct Ptracer {
     /// Ptrace options that will be applied to tracees, by default.
@@ -193,13 +279,15 @@ impl Ptracer {
                 ptrace::syscall(pid, pending),
         };
 
-        r.map_err(|source| Error::Restart {
-            pid,
-            mode: restart,
-            source,
-        })
+        r.died_if_esrch(pid)?;
+
+        Ok(())
     }
 
+    /// Spawn `cmd` for tracing.
+    ///
+    /// The command will be configured to requrest `PTRACE_TRACEME` after `fork()` and
+    /// pre-`exec()`. The caller will use this to avoid races and missed events.
     pub fn spawn(&mut self, mut cmd: Command) -> Result<Child> {
         // On fork, request `PTRACE_TRACEME`.
         unsafe {
@@ -217,10 +305,10 @@ impl Ptracer {
         Ok(child)
     }
 
-    /// Attach to a running tracee. This will deliver a SIGSTOP.
+    /// Attach to a running tracee. This will deliver a `SIGSTOP`.
     ///
-    /// Warning: the tracee may not be considered stopped until it has been seen
-    /// to stop via `wait()`.
+    /// **Warning:** the tracee may not be considered stopped until it has been seen to
+    /// stop via `wait()`.
     pub fn attach(&mut self, pid: Pid) -> Result<()> {
         let r = ptrace::attach(pid);
         let r = r.map_err(|source| Error::Attach { pid, source });
@@ -231,6 +319,8 @@ impl Ptracer {
     }
 
     /// Wait for some running tracee process to stop.
+    ///
+    /// If there are no tracees to wait on, returns `None`.
     pub fn wait(&mut self) -> Result<Option<Tracee>> {
         use Signal::*;
 
@@ -268,7 +358,7 @@ impl Ptracer {
                     // `PTRACE_O_TRACEEXEC` is not set by default, so it is not set when the child
                     // requests the attach. We will thus see its exec as a `SIGTRAP`, no matter what
                     // is set in `self.options`.
-                    let stop = Stop::SyscallExitStop(pid);
+                    let stop = Stop::SyscallExit;
                     let mut tracee = Tracee::new(pid, None, stop);
 
                     // Update the tracee state so subsequent traps are interpreted correctly.
@@ -280,16 +370,16 @@ impl Ptracer {
 
                     tracee
                 } else {
-                    let stop = Stop::SignalDeliveryStop(pid, SIGTRAP);
+                    let stop = Stop::SignalDelivery { signal: SIGTRAP };
                     Tracee::new(pid, None, stop)
                 }
             },
-            WaitStatus::Stopped(pid, sig) => {
-                if sig == SIGSTOP {
+            WaitStatus::Stopped(pid, signal) => {
+                if signal == SIGSTOP {
                     if let Some(state) = self.tracee_state_mut(pid) {
                         if *state == State::Attaching {
                             *state = State::Traced;
-                            let stop = Stop::AttachStop(pid);
+                            let stop = Stop::Attach;
                             let tracee = Tracee::new(pid, None, stop);
                             return Ok(Some(tracee));
                         }
@@ -300,41 +390,43 @@ impl Ptracer {
                         // only exists to let us know that the next stop (i.e. this stop) is an
                         // attach-stop, we can directly initialize this tracee as `Traced`.
                         self.set_tracee_state(pid, State::Traced);
-                        let stop = Stop::AttachStop(pid);
+                        let stop = Stop::Attach;
                         let tracee = Tracee::new(pid, None, stop);
                         return Ok(Some(tracee));
                     }
                 }
 
-                let stop = if is_group_stop(pid, sig)? {
-                    Stop::GroupStop(pid, sig)
+                let stop = if is_group_stop(pid, signal)? {
+                    Stop::Group { signal }
                 } else {
-                    Stop::SignalDeliveryStop(pid, sig)
+                    Stop::SignalDelivery { signal }
                 };
 
-                Tracee::new(pid, sig, stop)
+                Tracee::new(pid, signal, stop)
             },
-            WaitStatus::PtraceEvent(pid, sig, code) => {
+            WaitStatus::PtraceEvent(pid, signal, code) => {
                 match code {
                     libc::PTRACE_EVENT_FORK => {
-                        let new_pid = Pid::from_raw(ptrace::getevent(pid)? as u32 as i32);
+                        let evt_data = ptrace::getevent(pid).died_if_esrch(pid)?;
+                        let new = Pid::from_raw(evt_data as u32 as i32);
 
-                        // When we return, `new_pid` will start as a tracee, but will be delivered
-                        // a `SIGSTOP`. Mark it so we can recognize the `SIGSTOP` as an attach-stop.
-                        self.mark_tracee(new_pid);
+                        // When we return, `new` will start as a tracee, but will be delivered a
+                        // `SIGSTOP`. Mark it so we can recognize the `SIGSTOP` as an attach-stop.
+                        self.mark_tracee(new);
 
-                        let stop = Stop::Fork(pid, new_pid);
-                        Tracee::new(pid, sig, stop)
+                        let stop = Stop::Fork { new };
+                        Tracee::new(pid, signal, stop)
                     },
                     libc::PTRACE_EVENT_CLONE => {
-                        let new_pid = Pid::from_raw(ptrace::getevent(pid)? as u32 as i32);
+                        let evt_data = ptrace::getevent(pid).died_if_esrch(pid)?;
+                        let new = Pid::from_raw(evt_data as u32 as i32);
 
-                        // When we return, `new_pid` will start as a tracee, but will be delivered
-                        // a `SIGSTOP`. Mark it so we can recognize the `SIGSTOP` as an attach-stop.
-                        self.mark_tracee(new_pid);
+                        // When we return, `new` will start as a tracee, but will be delivered a
+                        // `SIGSTOP`. Mark it so we can recognize the `SIGSTOP` as an attach-stop.
+                        self.mark_tracee(new);
 
-                        let stop = Stop::Clone(pid, new_pid);
-                        Tracee::new(pid, sig, stop)
+                        let stop = Stop::Clone { new };
+                        Tracee::new(pid, signal, stop)
                     },
                     libc::PTRACE_EVENT_EXEC => {
                         // We are in one of two cases. The exec has either occurred on the main
@@ -342,60 +434,63 @@ impl Ptracer {
                         // execing thread will be equal to the tgid. In the off-main case, this is
                         // a change, and the old state for the tid == tgid will be invalid.
 
-                        // The current `pid` is now equal to the tgid of `old_pid`.
-                        let old_pid = Pid::from_raw(ptrace::getevent(pid)? as u32 as i32);
+                        // The current `pid` is now equal to the tgid of `old`.
+                        let evt_data = ptrace::getevent(pid).died_if_esrch(pid)?;
+                        let old = Pid::from_raw(evt_data as u32 as i32);
 
-                        if old_pid != pid {
+                        if old != pid {
                             // We exec'd off-thread, and previous tid state is now invalid.
-                            self.remove_tracee(old_pid);
+                            self.remove_tracee(old);
                         }
 
                         // We know we are in a syscall. Make sure we can correctly label the next
                         // syscall-stop as an exit-stop.
                         //
                         // Important: if we trace all the syscall-stops, we will report the syscall-
-                        // enter-stop as occurring on `old_pid`, but its matching syscall-exit-stop
-                        // as occurring on `pid`. This is correct, but might look odd.
+                        // enter-stop as occurring on `old`, but its matching syscall-exit-stop as
+                        // occurring on `pid`. This is correct, but might look odd.
                         self.set_tracee_state(pid, State::Syscalling);
 
-                        let stop = Stop::Exec(old_pid, pid);
+                        let stop = Stop::Exec { old };
 
-                        Tracee::new(pid, sig, stop)
+                        Tracee::new(pid, signal, stop)
                     },
                     libc::PTRACE_EVENT_EXIT => {
                         // In this context, `PTRACE_GETEVENTMSG` returns the pending wait status
                         // as an `unsigned long`. We are only interested in the low 16-bit word.
-                        let status = ptrace::getevent(pid)? as u16;
+                        let status = ptrace::getevent(pid).died_if_esrch(pid)? as u16;
 
                         self.remove_tracee(pid);
 
                         let stop = match ExitType::parse(status)? {
                             ExitType::Exit(exit_code) =>
-                                Stop::Exiting(pid, exit_code),
-                            ExitType::Signaled(sig, core_dumped) =>
-                                Stop::Signaling(pid, sig, core_dumped),
+                                Stop::Exiting { exit_code },
+                            ExitType::Signaled(signal, core_dumped) =>
+                                Stop::Signaling { signal, core_dumped },
                         };
 
-                        Tracee::new(pid, sig, stop)
+                        Tracee::new(pid, signal, stop)
                     },
                     libc::PTRACE_EVENT_VFORK => {
-                        let new_pid = Pid::from_raw(ptrace::getevent(pid)? as u32 as i32);
-                        self.mark_tracee(new_pid);
+                        let evt_data = ptrace::getevent(pid).died_if_esrch(pid)?;
+                        let new = Pid::from_raw(evt_data as u32 as i32);
+                        self.mark_tracee(new);
 
-                        let stop = Stop::Vfork(pid, new_pid);
+                        let stop = Stop::Vfork { new };
 
-                        Tracee::new(pid, sig, stop)
+                        Tracee::new(pid, signal, stop)
                     },
                     libc::PTRACE_EVENT_VFORK_DONE => {
-                        let new_pid = Pid::from_raw(ptrace::getevent(pid)? as u32 as i32);
-                        let stop = Stop::VforkDone(pid, new_pid);
+                        let evt_data = ptrace::getevent(pid).died_if_esrch(pid)?;
+                        let new = Pid::from_raw(evt_data as u32 as i32);
+                        let stop = Stop::VforkDone {new };
 
-                        Tracee::new(pid, sig, stop)
+                        Tracee::new(pid, signal, stop)
                     },
                     libc::PTRACE_EVENT_SECCOMP => {
                         // `SECCOMP_RET_DATA`, which is the low 16 bits of an int.
-                        let ret_data = ptrace::getevent(pid)? as u16;
-                        let stop = Stop::Seccomp(ret_data);
+                        let data = ptrace::getevent(pid).died_if_esrch(pid)? as u16;
+                        let stop = Stop::Seccomp { data };
 
                         if let Some(state) = self.tracee_state_mut(pid) {
                             *state = State::Syscalling;
@@ -403,7 +498,7 @@ impl Ptracer {
                             internal_error!("seccomp ptrace-event-stop for non-tracee");
                         }
 
-                        Tracee::new(pid, sig, stop)
+                        Tracee::new(pid, signal, stop)
                     },
                     libc::PTRACE_EVENT_STOP => {
                         // Unreachable by us, since we do not expose `PTRACE_SEIZE` &c.
@@ -448,11 +543,11 @@ impl Ptracer {
                         match state {
                             State::Syscalling => {
                                 *state = State::Traced;
-                                Stop::SyscallExitStop(pid)
+                                Stop::SyscallExit
                             },
                             State::Traced => {
                                 *state = State::Syscalling;
-                                Stop::SyscallEnterStop(pid)
+                                Stop::SyscallEnter
                             },
                             State::Attaching => {
                                 // A tracee in this state is waiting for a `SIGSTOP`, which is an
